@@ -1,46 +1,40 @@
-/* tn5250 -- an implentation of the 5250 telnet protocol.
+/* TN5250 - An implementation of the 5250 telnet protocol.
  * Copyright (C) 1997 Michael Madore
- *
+ * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
+ * the Free Software Foundation; either version 2, or (at your option)
+ * any later version.
+ * 
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- *
+ * 
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
- */
-
-#include "config.h"
+ * along with this software; see the file COPYING.  If not, write to
+ * the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
+ * Boston, MA 02111-1307 USA
+ * 
+ * As a special exception, the copyright holder gives permission
+ * for additional uses of the text contained in its release of TN5250.
+ * 
+ * The exception is that, if you link the TN5250 library with other files
+ * to produce an executable, this does not by itself cause the
+ * resulting executable to be covered by the GNU General Public License.
+ * Your use of that executable is in no way restricted on account of
+ * linking the TN5250 library code into it.
+ * 
+ * This exception does not however invalidate any other reasons why
+ * the executable file might be covered by the GNU General Public License.
+ * 
+ * If you write modifications of your own for TN5250, it is your choice
+ * whether to permit this exception to apply to your modifications.
+ * If you do not wish that, delete this exception notice. */
+#define _TN5250_TERMINAL_PRIVATE_DEFINED
+#include "tn5250-private.h"
 
 #if USE_SLANG
-#define _TN5250_TERMINAL_PRIVATE_DEFINED
-
-#include <stdio.h>
-#include <sys/time.h>
-#include <signal.h>
-#include <string.h>
-#include <ctype.h>
-#include <malloc.h>
-#ifdef HAVE_SLANG_H
-#include <slang.h>
-#else
-#ifdef HAVE_SLANG_SLANG_H
-#include <slang/slang.h>
-#endif
-#endif
-#include "utility.h"
-#include "dbuffer.h"
-#include "buffer.h"
-#include "record.h"
-#include "stream.h"
-#include "terminal.h"
-#include "slangterm.h"
 
 /* Mapping of 5250 colors to curses colors */
 #define A_5250_WHITE    0x100
@@ -100,11 +94,12 @@ static int slang_terminal_width(Tn5250Terminal * This);
 static int slang_terminal_height(Tn5250Terminal * This);
 static int slang_terminal_flags(Tn5250Terminal * This);
 static void slang_terminal_update(Tn5250Terminal * This,
-				   Tn5250DBuffer * dsp);
+				   Tn5250Display * display);
 static void slang_terminal_update_indicators(Tn5250Terminal * This,
-					      Tn5250DBuffer * dsp);
+					      Tn5250Display *display);
 static int slang_terminal_waitevent(Tn5250Terminal * This);
 static int slang_terminal_getkey(Tn5250Terminal * This);
+static void slang_terminal_beep(Tn5250Terminal * This);
 static int slang_terminal_get_esc_key(Tn5250Terminal * This, int is_esc);
 static void slang_terminal_set_attrs (Tn5250Terminal * This, int attrs);
 
@@ -114,38 +109,58 @@ struct _Tn5250TerminalPrivate {
    int attrs;
 };
 
-/*
+/****f* lib5250/tn5250_slang_terminal_new
+ * NAME
+ *    tn5250_slang_terminal_new
+ * SYNOPSIS
+ *    ret = tn5250_slang_terminal_new ();
+ * INPUTS
+ *    None
+ * DESCRIPTION
  *    Create a new curses terminal object.
- */
+ *****/
 Tn5250Terminal *tn5250_slang_terminal_new()
 {
    Tn5250Terminal *r = tn5250_new(Tn5250Terminal, 1);
-   if (r != NULL) {
-      r->data = tn5250_new(struct _Tn5250TerminalPrivate, 1);
-      if (r->data == NULL) {
-	 free(r);
-	 return NULL;
-      }
-      r->data->quit_flag = 0;
-      r->data->last_width = 0;
-      r->data->last_height = 0;
-      r->data->attrs = 0;
+   if (r == NULL)
+      return NULL;
 
-      r->conn_fd = -1;
-      r->init = slang_terminal_init;
-      r->term = slang_terminal_term;
-      r->destroy = slang_terminal_destroy;
-      r->width = slang_terminal_width;
-      r->height = slang_terminal_height;
-      r->flags = slang_terminal_flags;
-      r->update = slang_terminal_update;
-      r->update_indicators = slang_terminal_update_indicators;
-      r->waitevent = slang_terminal_waitevent;
-      r->getkey = slang_terminal_getkey;
+   r->data = tn5250_new(struct _Tn5250TerminalPrivate, 1);
+   if (r->data == NULL) {
+      free(r);
+      return NULL;
    }
+
+   r->data->quit_flag = 0;
+   r->data->last_width = 0;
+   r->data->last_height = 0;
+   r->data->attrs = 0;
+
+   r->conn_fd = -1;
+   r->init = slang_terminal_init;
+   r->term = slang_terminal_term;
+   r->destroy = slang_terminal_destroy;
+   r->width = slang_terminal_width;
+   r->height = slang_terminal_height;
+   r->flags = slang_terminal_flags;
+   r->update = slang_terminal_update;
+   r->update_indicators = slang_terminal_update_indicators;
+   r->waitevent = slang_terminal_waitevent;
+   r->getkey = slang_terminal_getkey;
+   r->beep = slang_terminal_beep;
    return r;
 }
 
+/****i* lib5250/slang_terminal_init
+ * NAME
+ *    slang_terminal_init
+ * SYNOPSIS
+ *    slang_terminal_init (This);
+ * INPUTS
+ *    Tn5250Terminal *     This       - 
+ * DESCRIPTION
+ *    DOCUMENT ME!!!
+ *****/
 static void slang_terminal_init(Tn5250Terminal * This)
 {
    SLtt_get_terminfo ();
@@ -177,12 +192,32 @@ static void slang_terminal_init(Tn5250Terminal * This)
    This->data->quit_flag = 0;
 }
 
+/****i* lib5250/slang_terminal_term
+ * NAME
+ *    slang_terminal_term
+ * SYNOPSIS
+ *    slang_terminal_term (This);
+ * INPUTS
+ *    Tn5250Terminal *     This       - 
+ * DESCRIPTION
+ *    DOCUMENT ME!!!
+ *****/
 static void slang_terminal_term(Tn5250Terminal * This)
 {
    SLsmg_reset_smg ();
    SLang_reset_tty ();
 }
 
+/****i* lib5250/slang_terminal_destroy
+ * NAME
+ *    slang_terminal_destroy
+ * SYNOPSIS
+ *    slang_terminal_destroy (This);
+ * INPUTS
+ *    Tn5250Terminal *     This       - 
+ * DESCRIPTION
+ *    DOCUMENT ME!!!
+ *****/
 static void slang_terminal_destroy(Tn5250Terminal * This)
 {
    if (This->data != NULL)
@@ -190,18 +225,48 @@ static void slang_terminal_destroy(Tn5250Terminal * This)
    free(This);
 }
 
+/****i* lib5250/slang_terminal_width
+ * NAME
+ *    slang_terminal_width
+ * SYNOPSIS
+ *    ret = slang_terminal_width (This);
+ * INPUTS
+ *    Tn5250Terminal  *    This       - 
+ * DESCRIPTION
+ *    DOCUMENT ME!!!
+ *****/
 static int slang_terminal_width(Tn5250Terminal /*@unused@*/ * This)
 {
    SLtt_get_screen_size();
    return SLtt_Screen_Cols;
 }
 
+/****i* lib5250/slang_terminal_height
+ * NAME
+ *    slang_terminal_height
+ * SYNOPSIS
+ *    ret = slang_terminal_height (This);
+ * INPUTS
+ *    Tn5250Terminal  *    This       - 
+ * DESCRIPTION
+ *    DOCUMENT ME!!!
+ *****/
 static int slang_terminal_height(Tn5250Terminal /*@unused@*/ * This)
 {
    SLtt_get_screen_size();
    return SLtt_Screen_Rows;
 }
 
+/****i* lib5250/slang_terminal_flags
+ * NAME
+ *    slang_terminal_flags
+ * SYNOPSIS
+ *    ret = slang_terminal_flags (This);
+ * INPUTS
+ *    Tn5250Terminal  *    This       - 
+ * DESCRIPTION
+ *    DOCUMENT ME!!!
+ *****/
 static int slang_terminal_flags(Tn5250Terminal /*@unused@*/ * This)
 {
    int f = 0;
@@ -210,30 +275,41 @@ static int slang_terminal_flags(Tn5250Terminal /*@unused@*/ * This)
    return f;
 }
 
-static void slang_terminal_update(Tn5250Terminal * This, Tn5250DBuffer * dsp)
+/****i* lib5250/slang_terminal_update
+ * NAME
+ *    slang_terminal_update
+ * SYNOPSIS
+ *    slang_terminal_update (This, display);
+ * INPUTS
+ *    Tn5250Terminal *     This       - 
+ *    Tn5250Display *      display    - 
+ * DESCRIPTION
+ *    DOCUMENT ME!!!
+ *****/
+static void slang_terminal_update(Tn5250Terminal * This, Tn5250Display * display)
 {
    int my, mx;
    int y, x;
    int curs_attr, temp_attr;
    unsigned char a = 0x20, c;
 
-   if (This->data->last_width != tn5250_dbuffer_width(dsp)
-       || This->data->last_height != tn5250_dbuffer_height(dsp)) {
+   if (This->data->last_width != tn5250_display_width(display)
+       || This->data->last_height != tn5250_display_height(display)) {
       SLsmg_cls ();
-      This->data->last_width = tn5250_dbuffer_width(dsp);
-      This->data->last_height = tn5250_dbuffer_height(dsp);
-      slang_terminal_update_indicators(This, dsp);
+      This->data->last_width = tn5250_display_width(display);
+      This->data->last_height = tn5250_display_height(display);
+      slang_terminal_update_indicators(This, display);
    }
    SLsmg_normal_video ();
    my = SLtt_Screen_Rows - 1;
    mx = SLtt_Screen_Cols - 1;
-   for (y = 0; y < tn5250_dbuffer_height(dsp); y++) {
+   for (y = 0; y < tn5250_display_height(display); y++) {
       if (y > my)
 	 break;
 
       SLsmg_gotorc (y, 0);
-      for (x = 0; x < tn5250_dbuffer_width(dsp); x++) {
-	 c = tn5250_dbuffer_char_at(dsp, y, x);
+      for (x = 0; x < tn5250_display_width(display); x++) {
+	 c = tn5250_display_char_at(display, y, x);
 	 if ((c & 0xe0) == 0x20) {	/* ATTRIBUTE */
 	    a = (c & 0xff);
 	    temp_attr = This->data->attrs;
@@ -248,7 +324,7 @@ static void slang_terminal_update(Tn5250Terminal * This, Tn5250DBuffer * dsp)
 	       SLsmg_write_char (' ');
 	       slang_terminal_set_attrs (This, temp_attr);
 	    } else {
-	       c = tn5250_ebcdic2ascii(c);
+	       c = tn5250_char_map_to_local (tn5250_display_char_map (display), c);
 	       if ((c < 0x20 && c > 0x00) || c >= 0x7f) { /* UNPRINTABLE */
 		  c = ' ';
 		  curs_attr ^= A_REVERSE;
@@ -271,13 +347,24 @@ static void slang_terminal_update(Tn5250Terminal * This, Tn5250DBuffer * dsp)
       }				/* for (int x ... */
    }				/* for (int y ... */
 
-   SLsmg_gotorc (tn5250_dbuffer_cursor_y(dsp), tn5250_dbuffer_cursor_x(dsp));
+   SLsmg_gotorc (tn5250_display_cursor_y(display), tn5250_display_cursor_x(display));
    SLsmg_refresh();
 }
 
-static void slang_terminal_update_indicators(Tn5250Terminal * This, Tn5250DBuffer * dsp)
+/****i* lib5250/slang_terminal_update_indicators
+ * NAME
+ *    slang_terminal_update_indicators
+ * SYNOPSIS
+ *    slang_terminal_update_indicators (This, display);
+ * INPUTS
+ *    Tn5250Terminal *     This       - 
+ *    Tn5250Display *      display    - 
+ * DESCRIPTION
+ *    DOCUMENT ME!!!
+ *****/
+static void slang_terminal_update_indicators(Tn5250Terminal * This, Tn5250Display * display)
 {
-   int inds = tn5250_dbuffer_indicators(dsp);
+   int inds = tn5250_display_indicators(display);
    char ind_buf[80];
 
    memset(ind_buf, ' ', sizeof(ind_buf));
@@ -296,14 +383,25 @@ static void slang_terminal_update_indicators(Tn5250Terminal * This, Tn5250DBuffe
       memcpy(ind_buf + 30, "IM", 2);
    }
    SLsmg_normal_video ();
-   SLsmg_gotorc (tn5250_dbuffer_height (dsp), 0);
+   SLsmg_gotorc (tn5250_display_height (display), 0);
    SLsmg_write_nchars (ind_buf, 80);
-   SLsmg_gotorc (tn5250_dbuffer_cursor_y(dsp), tn5250_dbuffer_cursor_x(dsp));
+   SLsmg_gotorc (tn5250_display_cursor_y(display), tn5250_display_cursor_x(display));
    SLsmg_refresh();
 }
 
+/****i* lib5250/slang_terminal_waitevent
+ * NAME
+ *    slang_terminal_waitevent
+ * SYNOPSIS
+ *    ret = slang_terminal_waitevent (This);
+ * INPUTS
+ *    Tn5250Terminal *     This       - 
+ * DESCRIPTION
+ *    DOCUMENT ME!!!
+ *****/
 static int slang_terminal_waitevent(Tn5250Terminal * This)
 {
+#if !defined(WIN32) && !defined(WINE)
    fd_set fdr;
    int result = 0;
    int sm;
@@ -330,8 +428,21 @@ static int slang_terminal_waitevent(Tn5250Terminal * This)
       result |= TN5250_TERMINAL_EVENT_DATA;
 
    return result;
+#else
+   /* XXX: WIN32/WINE - need to do this using WaitForMultipleObjects */
+#endif
 }
 
+/****i* lib5250/slang_terminal_getkey
+ * NAME
+ *    slang_terminal_getkey
+ * SYNOPSIS
+ *    ret = slang_terminal_getkey (This);
+ * INPUTS
+ *    Tn5250Terminal *     This       - 
+ * DESCRIPTION
+ *    DOCUMENT ME!!!
+ *****/
 static int slang_terminal_getkey(Tn5250Terminal * This)
 {
    unsigned int key;
@@ -394,7 +505,7 @@ static int slang_terminal_getkey(Tn5250Terminal * This)
 	    This->data->quit_flag = 1;
 	 return -1;
 
-      case 127:
+      case SL_KEY_DELETE:
 	 return K_DELETE;
 
       case SL_KEY_F(1):
@@ -487,21 +598,41 @@ static int slang_terminal_getkey(Tn5250Terminal * This)
    }
 }
 
-/*
+/****i* lib5250/slang_terminal_beep
+ * NAME
+ *    slang_terminal_beep
+ * SYNOPSIS
+ *    slang_terminal_beep (This);
+ * INPUTS
+ *    Tn5250Terminal *     This       - 
+ * DESCRIPTION
+ *    DOCUMENT ME!!!
+ *****/
+static void slang_terminal_beep (Tn5250Terminal * This)
+{
+   SLtt_beep ();
+}
+
+/****i* lib5250/slang_terminal_get_esc_key
+ * NAME
+ *    slang_terminal_get_esc_key
+ * SYNOPSIS
+ *    ret = slang_terminal_get_esc_key (This, is_esc);
+ * INPUTS
+ *    Tn5250Terminal *     This       - 
+ *    int                  is_esc     - 
+ * DESCRIPTION
  *    If a vt100 escape key sequence was introduced (using either
- *      <Esc> or <Ctrl+g>), handle the next key in the sequence.
- */
-static int slang_terminal_get_esc_key(Tn5250Terminal * This,
-				       int is_esc)
+ *    <Esc> or <Ctrl+g>), handle the next key in the sequence.
+ *****/
+static int slang_terminal_get_esc_key(Tn5250Terminal * This, int is_esc)
 {
    int y, x, key, display_key;
-   fd_set fdr;
 
    y = SLsmg_get_row ();
    x = SLsmg_get_column ();
    SLsmg_normal_video ();
-   /* FIXME: This might need to be row 28 */
-   SLsmg_gotorc (24, 60);
+   SLsmg_gotorc (This->data->last_height, 60);
    if (is_esc)
       SLsmg_write_string ("Esc ");
    else
@@ -509,9 +640,16 @@ static int slang_terminal_get_esc_key(Tn5250Terminal * This,
    SLsmg_gotorc (y, x);
    SLsmg_refresh();
 
-   FD_ZERO(&fdr);
-   FD_SET(0, &fdr);
-   select(1, &fdr, NULL, NULL, NULL);
+#if !defined(WIN32) && !defined(WINE)
+   {
+      fd_set fdr;
+      FD_ZERO(&fdr);
+      FD_SET(0, &fdr);
+      select(1, &fdr, NULL, NULL, NULL);
+   }
+#else
+   /* XXX: What do we need that for? */
+#endif
    key = SLkp_getkey ();
 
    if (isalpha(key))
@@ -653,8 +791,7 @@ static int slang_terminal_get_esc_key(Tn5250Terminal * This,
       break;
    }
 
-   /* FIXME: This might need to be row 27 */
-   SLsmg_gotorc (24, 64);
+   SLsmg_gotorc (This->data->last_height, 64);
    if (key == SL_KEY_ERR)
       SLsmg_write_string ("???");
    else
@@ -664,6 +801,17 @@ static int slang_terminal_get_esc_key(Tn5250Terminal * This,
    return key;
 }
 
+/****i* lib5250/slang_terminal_set_attrs
+ * NAME
+ *    slang_terminal_set_attrs
+ * SYNOPSIS
+ *    slang_terminal_set_attrs (This, attrs);
+ * INPUTS
+ *    Tn5250Terminal *     This       - 
+ *    int                  attrs      - 
+ * DESCRIPTION
+ *    DOCUMENT ME!!!
+ *****/
 static void slang_terminal_set_attrs (Tn5250Terminal * This, int attrs)
 {
    if (attrs == This->data->attrs)
